#include "Vhub75e.h"
#include "tests.hpp"
#include "devices.hpp"
#include "verilated.h"
#include "verilated_vcd_c.h"
#include <array>
#include <catch2/catch_test_macros.hpp>
#include <memory>
#include <stdio.h>
#include <vector>

void LineDriverTest(VerilatedContext &ctx) {
  // create the hub75e driver and run some basic tests

  // we generate 512 random color values (24 bits)
  // we load them in.
  unsigned long posedge = 0; // counts positive edges.
  unsigned long simtime = 0;
  auto dut = std::make_unique<Vhub75e>(&ctx, "dut");

  VerilatedVcdC *m_trace = new VerilatedVcdC;
  dut->trace(m_trace, 5);

  auto bram = FakeBRAM(1, dut->clk, dut->pixbuf_addr, dut->pixbuf_data);

  auto hub75 = HUB75Reciever(128, 64, *dut);
  printf("Performing basic test\n");
  bool done = false;
  m_trace->open("waveform.vcd");

  while (!done) {
    dut->clk ^= 1; // toggle clock
    dut->eval();

    if (dut->clk == 1) {
      // rising edge.
      posedge++;
      bram.tick();
    }

    dut->write_trig = posedge < 2 ? 1 : 0;

    if (posedge >= 250000) {
      done = true;
    }
    hub75.tick();
    m_trace->dump(simtime);
    simtime++;
  }
  m_trace->close();
};

// int main() {
//   auto ctx = std::make_unique<VerilatedContext>();
//   ctx->traceEverOn(true);
//   printf("hello world!\n");
//   LineDriverTest(*ctx);
// };

TEST_CASE("Hub75 Test") {
  auto ctx = std::make_unique<VerilatedContext>();
  // setup DUT

  unsigned long posedge = 0; // counts positive edges.
  unsigned long simtime = 0;
  auto dut = std::make_unique<Vhub75e>(ctx.get(), "dut");

  auto bram = FakeBRAM(1, dut->clk, dut->pixbuf_addr, dut->pixbuf_data);

  auto hub75 = HUB75Reciever(128, 64, *dut);
  bool done = false;

  bool driver_done = false;
  while (!done) {
    dut->clk ^= 1; // toggle clock
    dut->eval();

    if (dut->clk == 1) {
      // rising edge.
      posedge++;
    }

    dut->write_trig = posedge < 2 ? 1 : 0;
    if (dut->done) {
      done = true;
      driver_done = true;
    }

    if (posedge >= 250000) {
      done = true;
      driver_done = false;
    }
    hub75.tick();
    bram.tick();
    simtime++;
  }
  REQUIRE(driver_done);
  SECTION("Bit Sizing/Count") {

    CHECK(hub75.get_past_rows().size() == 8);

    auto rows = hub75.get_past_rows();

    for (int i = 0; i < rows.size(); i++) {
      auto r = rows[i];
    }
  }
  SECTION("Pulse width") {}
}

TEST_CASE("HUB75E Driver Test") {
  auto fixture = VerilatorTestFixture<Vhub75e>();
  // very simple done checker.
  auto done_check = [](Vhub75e &dut, unsigned long time) {
    return dut.done == 1;
  };
  fixture.set_done_callback(done_check);
  const Vhub75e &dut = fixture.get();
  auto stim = std::make_shared<PulseStimulus>(dut.write_trig, 4);
  fixture.set_timeout(250000);

  fixture.add_module(stim);

  SECTION("Smoke Tests") {
    fixture.enable_trace("testing.vcd");
    auto bram = std::make_shared<FakeBRAM>(1, dut.clk, dut.pixbuf_addr,
                                           dut.pixbuf_data);
    fixture.add_module(bram);
    auto display = std::make_shared<HUB75Reciever>(128, 64, dut);
    fixture.add_module(display);

    fixture.exec();

    CHECK(fixture.get_reason() ==
          VerilatorTestFixture<Vhub75e>::FinishReason::Ok);

    auto rows = display->get_past_rows();
    CHECK(rows.size() == 8);
    for (int i = 0; i < rows.size(); i++) {
      auto &[r0, r1] = rows[i];
      CHECK(r0.size() == 128);
      CHECK(r1.size() == 128);
    }
    // pulse width smoke tests.
    auto pulses = display->get_pulse_widths();
    REQUIRE(pulses.size() == rows.size());
    for (int i = 1; i < pulses.size(); i++) {
      REQUIRE(pulses[i] == pulses[i - 1] / 2);
    }
    auto [row0, row1] = display->transpose();
    REQUIRE(row0.size() == 128);
    REQUIRE(row1.size() == 128);
    auto ram_ref = bram->get();

    CAPTURE(row0);
    CHECK(std::equal(ram_ref.begin(), ram_ref.begin() + 128, row0.begin(),
                     row0.end()));
  }
  SECTION("Line Correctness") {
    // this is the part where we validate that the line in = line out.
    // we have to generate different values since the
  }
}
